// vim:syntax=javascript

class OneParticleLevel {
    function OneParticleLevel(
        parent_config : Object,
        message_setup : Function,
        tip_engine : Function,
        ipx : Number, ipy : Number,
        ivx : Number, ivy : Number,
        charge : Number,
        gx : Number, gy : Number,
        g_rad : Number,
        g_vel : Number,
        max_fields : Number,
        obstacles : Array) {

        var parent_mc : MovieClip = parent_config.mc;
        var screen_width : Number = parent_config.screen_width;
        var screen_height : Number = parent_config.screen_height;

        var sim_name = "sim_"+parent_config.name_count+"_mc";

        var message_name = "message_"+parent_config.name_count+"_mc";
        if (message_setup != null) {
            parent_mc.createEmptyMovieClip(message_name, Potential.MESSAGE_DEPTH);
            message_setup(parent_mc[message_name],
                function () {
                    parent_mc[sim_name].enabled = true;
                }
            );
        }

        parent_mc.attachMovie("simulation", sim_name, Potential.SIMULATION_DEPTH,
            {parent_config:parent_config});

        parent_mc[sim_name].specificSetup = function () {
            var static_this = this;

            if (message_setup != null) {
                this.enabled = false;
            }

            this.placeSprite(gx, gy, Particle.PARTICLE_SIZE, (charge<0?"neg_goal":"pos_goal"), this.getNextHighestDepth());
            this.particle1 = this.placeParticle(charge, ipx, ipy, ivx, ivy, this.getNextHighestDepth());
            this.goal = {x:gx, y:gy};
            this.max_fields = max_fields;

            with (this.obstacles_mc) {
                var obstacle_count = obstacles.length;
                for (var i = 0; i < obstacle_count; i ++) {
                    var obs_points : Array = obstacles[i];
                    var obs_point_count = obs_points.length;

                    moveTo(obs_points[0].x, obs_points[0].y);
                    lineStyle(1, 0x000000);
                    beginFill(0x000000);

                    for (var j = 1; j < obs_point_count; j ++) {
                        lineTo(obs_points[j].x, obs_points[j].y);
                    }
                    endFill();
                }
            }

            this.resetGoals = function () : Void {
            };

            this.updateGoals = function () : Void {
                if (static_this.particle1.hitTest(static_this.goal, g_rad)) {
                    var vx : Number = static_this.particle1.vx;
                    var vy : Number = static_this.particle1.vy;
                    if (vx*vx+vy*vy < g_vel*g_vel) {
                        static_this.removeMovieClip();
                    }
                }
            };

            this.tipEngine = tip_engine;

            this.resetGoals();
        };
    }

    static function gen_message_setup (pc:Object, message:String, width:Number, height:Number) : Function {
        var box_x : Number = (pc.screen_width - width)/2;
        var box_y : Number = (pc.screen_height - height)/2;

        return function (mc : MovieClip, onDone : Function) {
            mc._alpha = 80;
            RoundRect.roundRect(mc, box_x, box_y, width, height, 15);
            var text1 = message;
            var msg1 = VeraText.create(mc, "msg1", text1, 18, box_x+10, box_y+10, width-20, height-20);
            mc.onPress = function () {
                mc.removeMovieClip();
            }
            mc.onUnload = function () {
                onDone();
            }
        };
    }

    static function gen_tip_engine_1_2 (which_axis : String) : Function {
        return function (is_running : Boolean, fields : /*Field*/ Array, particles : /*Particle*/ Array) : String {
            if (fields.length == 0) {
                return "Tip: Click and drag to create a force field.";
            } else {
                var p = particles[0];
                var f = fields[0];
                var force = f.computeForce(p)
                if (is_running) {
                    return "";
                }

                if (force[which_axis] <= 0) {
                    return "Tip: Remove an unwanted field with Backspace.";
                } else {
                    return "Tip: Press Enter to run the simulation.";
                }
            }
        };
    }
}
