// vim:syntax=javascript

// properties:
// parent_config - object with global configuration referring back to the parent
// specificSetup - set up level layout and goals (resetGoals, updateGoals)

class Simulation extends MovieClip {
    static var TRAIL_DEPTH : Number = 1;
    static var OBSTACLES_DEPTH : Number = 2;
    var parent_config : Object;
    var screen_width : Number, screen_height : Number;
    var specificSetup : Function;

    var particles : /*Particle*/ Array;
    var trail_mc : MovieClip;
    var obstacles_mc : MovieClip;

    var fields : /*Field*/ Array;
    var temp_field : Field;
    var temp_field_start : Object;
    var max_fields : Number;

    var sim_running : Boolean;
    var key_listener : Object;

    var resetGoals : Function;
    var updateGoals : Function;

    var tipEngine : Function;

    static var FIELD_SNAP_SIZE : Number = 20;
    static var PARTICLE_SIZE : Number = Particle.PARTICLE_SIZE;
    static var FIELD_POTENTIAL : Number = 10;
    
    function placeParticle (charge : Number, x : Number, y : Number, vx : Number, vy : Number, depth : Number) : Particle {
        var name = "particle_"+depth+"_mc";
        if (charge > 0) {
            attachMovie("pos", name, depth);
        } else {
            attachMovie("neg", name, depth);
        }

        var new_particle = new Particle(this[name], charge, x, y, vx, vy);

        particles.push(new_particle);

        return new_particle;
    }

    // a dumb sprite, like a goal
    function placeSprite(x : Number, y : Number, size : Number, mc_name : String, depth : Number) {
        var name = mc_name+"_"+depth+"_mc";

        attachMovie(mc_name, name, depth);
        this[name]._width = size;
        this[name]._height = size;
        this[name]._x = x - size/2;
        this[name]._y = y - size/2;
    }

    function moveParticle(p : Particle) {
        // TODO: proper integration?

        var ax = 0;
        var ay = 0;
        //trace(p);

        forAllFields( function(f : Field) {
            var force = f.computeForce(p);
            ax += force.fx;
            ay += force.fy;
        });
        //trace(p+"\naccel ("+ax+","+ay+")");
        var vx = p.vx + ax;
        var vy = p.vy + ay
        var x = p.x + vx;
        var y = p.y + vy;

        // trail is looking mighty ugly at low speeds with Adobe Flash
        /*
        var old_round_x = Math.round(p.x);
        var old_round_y = Math.round(p.y);
        var new_round_x = Math.round(x);
        var new_round_y = Math.round(y);
        if (new_round_x != old_round_x ||
            new_round_y != old_round_y) {
            trail_mc.moveTo(old_round_x, old_round_y);
            trail_mc.lineStyle(0, 0x00000);
            trail_mc.lineTo(new_round_x, new_round_y);
        }
        */

        p.update(x, y, vx, vy);
    }

    function forAllParticles(f : Function) {
        var particle_count = particles.length;

        for (var i = 0; i < particle_count; i++) {
            f(particles[i]);
        }
    }

    function checkCollision(p : Particle) : Boolean {
        if (p.x - Particle.PARTICLE_SIZE/2 < 0 ||
            p.x + Particle.PARTICLE_SIZE/2 >= screen_width ||
            p.y - Particle.PARTICLE_SIZE/2 < 0 ||
            p.y + Particle.PARTICLE_SIZE/2 >= screen_height) {
            return true;
        }

        // 4 corners
        if (obstacles_mc.hitTest(p.x,
                                 p.y - Particle.PARTICLE_SIZE/2,
                                 true) ||
            obstacles_mc.hitTest(p.x + Particle.PARTICLE_SIZE/2,
                                 p.y,
                                 true) ||
            obstacles_mc.hitTest(p.x - Particle.PARTICLE_SIZE/2,
                                 p.y,
                                 true) ||
            obstacles_mc.hitTest(p.x,
                                 p.y + Particle.PARTICLE_SIZE/2,
                                 true)) {
            return true;
        }

        return false;
    }

    function makeField(low_x : Number, high_x : Number, low_y : Number, high_y : Number, magnitude : Number) : Field {
        var depth = getNextHighestDepth();
        var name = "field_"+depth+"_mc";
        attachMovie("field", name, depth);

        var new_field = new Field(this[name], low_x, high_x, low_y, high_y, screen_width, screen_height, magnitude);

        return new_field;
    }

    function placeField(low_x : Number, high_x : Number, low_y : Number, high_y : Number, magnitude : Number) {
        fields.push(makeField(low_x,high_x,low_y,high_y,magnitude));
    }

    function forAllFields(f : Function) {
        var field_count = fields.length;

        for (var i = 0; i < field_count; i++) {
            f(fields[i]);
        }
    }

    function startSim() {
        if (temp_field != null) {
            temp_field.remove();
            temp_field = null;
        }
        resetGoals();
        sim_running = true;
    }

    function endSim() {
        trail_mc.clear();

        forAllParticles( function(p : Particle) {
            p.reset();
        });
        sim_running = false;
    }

    function doStatus() {
        parent_config.reportStatus(sim_running , particles, fields, max_fields);
    }

    function doTips() {
        if (tipEngine != null) {
            parent_config.reportTip(tipEngine(sim_running, fields, particles));
        } else {
            parent_config.reportTip("");
        }
    }

    function onLoad() {
        var static_this = this;

        sim_running = false;

        particles = new Array();
        fields = new Array();
        temp_field = null;
        max_fields = 0;
        tipEngine = null;

        screen_width = parent_config.screen_width;
        screen_height = parent_config.screen_height;

        createEmptyMovieClip("trail_mc", TRAIL_DEPTH);
        createEmptyMovieClip("obstacles_mc", OBSTACLES_DEPTH);

        specificSetup();

        doStatus();
        doTips();

        key_listener = {
            onKeyDown : function () {
                var k:Number = Key.getCode();

                if (k == Key.ENTER) {
                    with (static_this) {
                        if (sim_running) {
                            endSim();
                        } else {
                            startSim();
                        }
                    }
                }
                if (k == Key.BACKSPACE) {
                    with (static_this) {
                        if (!sim_running) {
                            if (fields.length > 0) {
                                var dead_field = fields.pop();
                                dead_field.remove();
                            }
                        }
                    }
                }

                static_this.doTips();
            }
        };

        Key.addListener(key_listener);
    }

    function snapFieldPointX(x : Number, rounder : Function) : Number {
        x = Math.max(0, Math.min(x, screen_width));
        return rounder(x / FIELD_SNAP_SIZE) * FIELD_SNAP_SIZE;
    }
    function snapFieldPointY(y : Number, rounder : Function) : Number {
        y = Math.max(0, Math.min(y, screen_height));
        return rounder(y / FIELD_SNAP_SIZE) * FIELD_SNAP_SIZE;
    }


    function onMouseDown() {
        if (!enabled) {
            return;
        }
        if (sim_running) {
            return;
        }
        if (max_fields != 0 && fields.length >= max_fields) {
            return;
        }
        if (temp_field == null) {
            temp_field_start = {x:_xmouse, y:_ymouse};
            var high_x : Number = snapFieldPointX(_xmouse, Math.round);
            var high_y : Number = snapFieldPointY(_ymouse, Math.round);
            var init_x : Number = _xmouse - high_x;
            if (init_x < 0) {
                init_x = -FIELD_SNAP_SIZE;
            } else {
                init_x = FIELD_SNAP_SIZE;
            }
            // start a new field
            temp_field = makeField(high_x+init_x, high_x, high_y, high_y, FIELD_POTENTIAL);
        }
    }

    function onMouseMove() {
        if (!enabled) {
            return;
        }

        if (sim_running) {
            return;
        }

        if (temp_field != null) {
            var high_x : Number;
            var high_y : Number;
            var low_x : Number;
            var low_y : Number;
            var delta_x = _xmouse - temp_field_start.x;
            var delta_y = _ymouse - temp_field_start.y;
            if (Math.abs(delta_x) > Math.abs(delta_y)) {
                if (delta_x > 0) {
                    low_x = snapFieldPointX(_xmouse, Math.ceil);
                    high_x = snapFieldPointX(temp_field_start.x, Math.floor);
                } else {
                    low_x = snapFieldPointX(_xmouse, Math.floor);
                    high_x = snapFieldPointX(temp_field_start.x, Math.ceil);
                }
                high_y = low_y = 0;
            } else {
                if (delta_y > 0) {
                    low_y = snapFieldPointY(_ymouse, Math.ceil);
                    high_y = snapFieldPointY(temp_field_start.y, Math.floor);
                } else {
                    low_y = snapFieldPointY(_ymouse, Math.floor);
                    high_y = snapFieldPointY(temp_field_start.y, Math.ceil);
                }
                high_x = low_x = 0;
            }
            if (low_x != high_x || low_y != high_y) {
                temp_field.update(low_x, high_x, low_y, high_y);
            }
        }
    }

    function onMouseUp() {
        if (!enabled) {
            return;
        }
        if (sim_running) {
            return;
        }

        if (temp_field != null) {
            fields.push(temp_field);
            temp_field = null;
        }

        doTips();
    }

    function onEnterFrame () {
        if (!enabled) {
            return;
        }

        // simulation step

        var static_this : Simulation = this;
        var collided = false;

        if (sim_running) {
            forAllParticles( function(p : Particle) {
                with (static_this) {
                    moveParticle(p);
                    if (checkCollision(p)) {
                        collided = true;
                    }

                }
            });

            updateGoals();

            doStatus();

            if (collided) {
                endSim();
            }
        } else {
            doStatus();
        }
    }

    function onUnload () {
        Key.removeListener(key_listener);
        parent_config.reportCompletion();
    }
}
